// GBAMP Skinner.cpp
//
/*
   GBAMP Skinner.exe
   Written by chishm aka Michael Chisholm.
   Compiles bitmaps into a BACKIMG file for use
   with the GBA Movie Player.
*/

/*
	Source Code Information, Licensing and Copyright:
	I hereby release this program and all source code into
	the public domain. If you wish to port the code or use
	it in your own applications this is completely okay.
	Just email me first at chishm@hotmail.com as a simple courtesy.
	This has been written using MS VC++ version 6.0.
	It uses windows APIs for ini access. If you wish to port it,
	you may need to write your own ini code. The bitmap code should
	be cross platform compatible.
*/

/* 
	2005-04-16 - Prerelease alpha
	 * Bitmaps implemented
	 * Ini not implemented
	2005-04-17 - Version 1.0
	 * Ini file now implemented
	 * First official release
	 * Compatible only with firmware that includes game function
	2005-04-18 - Version 1.01
	 * Changed exit codes to reflect errors that occur during skinning
	2005-05-01 - Version 1.10
	 * Palette checking included
	2005-05-25 - Version 1.20
	 * Can now create BACKIMG files for the old firmware (e10)
	2005-06-08 - Version 1.21
	 * Changed error level to 1 for palette mismatches
*/


#include "stdafx.h"	
#include <stdio.h>
#include <stdlib.h>

#include "INI.h"

#define NUM_BG 4
#define NUM_MAIN_ICON 6
#define NUM_FILE_ICON 2

#define NUM_BMPS 12
#define NUM_OFFSETS 7
#define NUM_TEXT_COLORS 4

#define BG_WIDTH 240
#define BG_HEIGHT 160

#define MAIN_ICON_WIDTH 64
#define MAIN_ICON_HEIGHT 64

#define FILE_ICON_WIDTH 16
#define FILE_ICON_HEIGHT 16

#define PALETTE_LENGTH 1024
#define HEAD_LENGTH 54
#define PALETTE_RESERVED 0xFC

typedef struct                       /**** BMP file header structure ****/
{
    unsigned short bfType;           /* Magic number for file */
    unsigned long  bfSize;           /* Size of file */
    unsigned long  bfReserved;		 /* Reserved */
    unsigned long  bfOffBits;        /* Offset to bitmap data */
} BMPFILEHEAD;

typedef struct                       /**** BMP file info structure ****/
{
    unsigned int   biSize;           /* Size of info header */
    int            biWidth;          /* Width of image */
    int            biHeight;         /* Height of image */
    unsigned short biPlanes;         /* Number of color planes */
    unsigned short biBitCount;       /* Number of bits per pixel */
    unsigned int   biCompression;    /* Type of compression to use */
    unsigned int   biSizeImage;      /* Size of image data */
    int            biXPelsPerMeter;  /* X pixels per meter */
    int            biYPelsPerMeter;  /* Y pixels per meter */
    unsigned int   biClrUsed;        /* Number of colors used */
    unsigned int   biClrImportant;   /* Number of important colors */
} BMPINFOHEAD;

typedef struct							/* Info about location of images */
{
	CString		filename;				/* Name of bitmap */
	int			xOffset;				/* Offset of image data from left of bitmap */
	int			yOffset;				/* " top of bitmap */
	int			width;					/* Width of image */
	int			height;					/* Height of image */
} INI_BMP;

typedef struct							/* Positions of icons on GBAMP */
{
	unsigned char	xOffset;			/* Offset from left of screen */
	unsigned char	yOffset;			/* Offset from top of screen */
} INI_OFFSET;

typedef struct							/* Color in RGBN format */
{
	unsigned char red;
	unsigned char green;
	unsigned char blue;
	unsigned char reserved;
} RGBN;

/*
 Load the pixels of a 256 color bitmap into a char array.
 Width and height are inputs to test if bitmap is the 
 size expected.
 */
unsigned char* LoadBmpData(INI_BMP bmp)
{
	FILE* BMPFile;
	unsigned char *BMPData;
	BMPFILEHEAD FileHead;
	BMPINFOHEAD InfoHead;
	
	/* This checks for the file */
	BMPFile = fopen(bmp.filename, "rb");
	
	if (BMPFile == NULL)
	{
		printf("Cannot open %s.\n", bmp.filename);
		return(NULL);
	}
	
	/* Read the header information */
	fread(&(FileHead.bfType), sizeof(FileHead.bfType), 1, BMPFile);
	fread(&(FileHead.bfSize), sizeof(FileHead.bfSize), 1, BMPFile);
	fread(&(FileHead.bfReserved), sizeof(FileHead.bfReserved), 1, BMPFile);
	fread(&(FileHead.bfOffBits), sizeof(FileHead.bfOffBits), 1, BMPFile);

	fread(&InfoHead, sizeof(InfoHead), 1, BMPFile);

	/* debugging info
		printf("%d %02X\n",sizeof(FileHead.bfType),FileHead.bfType);
		printf("%d %04X\n",sizeof(FileHead.bfSize),FileHead.bfSize);
		printf("%d %04X\n",sizeof(FileHead.bfReserved),FileHead.bfReserved);
		printf("%d %04X\n",sizeof(FileHead.bfOffBits),FileHead.bfOffBits);
	*/
		
	if (FileHead.bfType != 19778 || FileHead.bfReserved != 0)
	{
		/* Not a valid bitmap file - don't display */
		printf("%s is not a valid bitmap.\n", bmp.filename);
		fclose(BMPFile);
		return(NULL);
	}
	
	if (InfoHead.biCompression != 0)
	{
		/*Compressed file - don't display*/
		printf("%s is a compressed file.\n", bmp.filename);
		printf("%1X",InfoHead.biCompression);
		fclose(BMPFile);
		return(NULL);
	}
	
	if (InfoHead.biBitCount != 8)
	{
		/*If the file is other than 8-bit dont read.*/
		printf("%s is not an 8-bit bitmap.\n", bmp.filename);
		fclose(BMPFile);
		return(NULL);
	}
	
	if (InfoHead.biWidth < (bmp.width  + bmp.xOffset))
	{
		/*If its wrong dimensions dont load.*/
		printf("%s is too narrow to load bitmap from.\n", bmp.filename);
		fclose(BMPFile);
		return(NULL);
	}

	if (InfoHead.biHeight < (bmp.height + bmp.yOffset))
	{
		/*If its wrong dimensions dont load.*/
		printf("%s is too short to load bitmap from.\n", bmp.filename);
		fclose(BMPFile);
		return(NULL);
	}

	/*Skip the palette*/
	fseek(BMPFile, PALETTE_LENGTH, SEEK_CUR);
	
	BMPData = (unsigned char*) malloc(bmp.width * bmp.height);

	/* Read bmp in reverse line order, at specific offset */
	printf("Reading %s, offset (%d,%d), size %dx%d.\n",bmp.filename,bmp.xOffset,bmp.yOffset,bmp.width,bmp.height);

	for (int line = 0; line < bmp.height; line++){
		/* Jump to:     |End of header              |  +   | Y offset from end of file                                  | +  | x offset |      */
		fseek (BMPFile,(PALETTE_LENGTH + HEAD_LENGTH) + ((InfoHead.biHeight - bmp.yOffset - line - 1) * InfoHead.biWidth) + bmp.xOffset, SEEK_SET);
		fread (&BMPData[line * bmp.width], bmp.width, 1, BMPFile);
	}
	fclose(BMPFile);
	return (BMPData);
}

/*
 Load the palette of an 8 bit bitmap into a char array.
 */
unsigned char* LoadBmpPalette(const char *filename)
{
	FILE* BMPFile;
	unsigned char *BMPPalette;
	BMPFILEHEAD FileHead;
	BMPINFOHEAD InfoHead;
	
	/* This checks for the file */
	BMPFile = fopen(filename, "rb");
	
	if (BMPFile == NULL)
	{
		printf("Cannot open %s.\n", filename);
		return(NULL);
	}

	/* Read the header information */
	fread(&(FileHead.bfType), sizeof(FileHead.bfType), 1, BMPFile);
	fread(&(FileHead.bfSize), sizeof(FileHead.bfSize), 1, BMPFile);
	fread(&(FileHead.bfReserved), sizeof(FileHead.bfReserved), 1, BMPFile);
	fread(&(FileHead.bfOffBits), sizeof(FileHead.bfOffBits), 1, BMPFile);

	fread(&InfoHead, sizeof(InfoHead), 1, BMPFile);

	if (FileHead.bfType != 19778 || FileHead.bfReserved != 0)
	{
		/* Not a valid bitmap file - don't display */
		printf("%s is not a valid bitmap.\n", filename);
		fclose(BMPFile);
		return(NULL);
	}
	
	if (InfoHead.biCompression != 0)
	{
		/*Compressed file - don't display*/
		printf("%s is a compressed file.\n", filename);
		printf("%1X",InfoHead.biCompression);
		fclose(BMPFile);
		return(NULL);
	}
	
	if (InfoHead.biBitCount != 8)
	{
		/*If the file is other than 8-bit dont read.*/
		printf("%s is not an 8-bit bitmap.\n", filename);
		fclose(BMPFile);
		return(NULL);
	}
	
			
	BMPPalette = (unsigned char*) malloc(PALETTE_LENGTH);

	printf("Reading palette from %s.\n",filename);

	/* Read bmp palette */
	fread(BMPPalette,PALETTE_LENGTH,1,BMPFile);

	fclose(BMPFile);
	return (BMPPalette);
}

/*
 Compare one bitmap's palette to another, return true if they match
 */
bool CompareBmpPalette(const char *file1, const char *file2)
{
	unsigned char* pal1;
	unsigned char* pal2;
	int i;
	
	// They're the same file
	if (strcmp(file1,file2) == 0) {
		return (true);
	}
	
	printf("Comparing palettes in %s and %s.\n", file1, file2);
	
	pal1 = LoadBmpPalette(file1);
	if (pal1 == NULL){
		return (false);
	}
	pal2 = LoadBmpPalette(file2);
	if (pal2 == NULL){
		return (false);
	}

	for ( i = 0; i < PALETTE_LENGTH; i++ ){
		if ( pal1[i] != pal2[i]) {
			printf ("%s has a different palette to %s in entry %d.\n", file1, file2, i);
			free(pal1);
			free(pal2);
			return (false);
		}
	}
	free(pal1);
	free(pal2);
	return (true);
}

/*
	Read bmp info from ini and return in an INI_BMP struct
*/
INI_BMP iniGetBmpInfo ( char* key, char* section, CINI &ini )
{
	INI_BMP bmpInfo;
	CString temp1 = ini.getKeyValue(key, section);
	CString temp2;
	// Goes through and extracts data from each comma delimited section in value
	temp2 = temp1.SpanExcluding(",");
	temp1.Delete(0,temp2.GetLength());
	temp2.TrimLeft("\" \t,");
	temp2.TrimRight("\" \t,");
	bmpInfo.filename = temp2;
	temp1.TrimLeft(", ");
	bmpInfo.xOffset=atoi(temp1);
	temp1.TrimLeft("0123456789.");
	temp1.TrimLeft(", ");
	bmpInfo.yOffset = atoi (temp1);
	temp1.TrimLeft("0123456789.");
	temp1.TrimLeft(", ");
	bmpInfo.width = atoi (temp1);
	temp1.TrimLeft("0123456789.");
	temp1.TrimLeft(", ");
	bmpInfo.height = atoi (temp1);

	return (bmpInfo);
}

/*
	Read Icon offsets from ini and return in INI_OFFSET struct
*/
INI_OFFSET iniGetOffset (char* key, char* section, CINI &ini )
{
	INI_OFFSET iconOffset;
	CString temp = ini.getKeyValue(key, section);
	temp.MakeLower();
	// Remove non numeric characters
	temp.TrimLeft("\" \t,abcdefghijklmnopqrstuvwxyz");
	iconOffset.xOffset = atoi(temp);
	// Remove first set of data
	temp.TrimLeft("0123456789.");
	temp.TrimLeft("\" \t,abcdefghijklmnopqrstuvwxyz");
	iconOffset.yOffset = atoi(temp);
	return (iconOffset);
}

/*
	Read Color triplets from ini file and return as a red, green, 
	blue, none value
*/
RGBN iniGetRGB (char* key, char* section, CINI &ini)
{
	RGBN color;
	CString temp = ini.getKeyValue(key, section);
	temp.MakeLower();
	// Remove non numeric characters
	temp.TrimLeft("\" \t,abcdefghijklmnopqrstuvwxyz");
	color.red = atoi(temp);
	// Remove leading set of data
	temp.TrimLeft("0123456789.");
	temp.TrimLeft("\" \t,abcdefghijklmnopqrstuvwxyz");
	color.green = atoi(temp);
	// Remove leading set of data
	temp.TrimLeft("0123456789.");
	temp.TrimLeft("\" \t,abcdefghijklmnopqrstuvwxyz");
	color.blue = atoi(temp);
	color.reserved = 0;

	return (color);
}

/*
	Main - GBAMP Skinner
	Converts bitmaps to correct format for use as a 
	skin on the GBAMP
*/
int main(int argc, char* argv[])
{
	// Program Version
	char* firmwareVersion = "e17";
	char* oldFirmwareVersion = "e10";
	int   iniVersion = 1;

	// Image names in ini file
	char* bmps[NUM_BMPS] = {"Main_BG","Main_Icon_Movie","Main_Icon_Music",
		"Main_Icon_Image","Main_Icon_Game","Main_Icon_EBook",
		"Main_Icon_Help","File_BG","File_Icon_Folder","File_Icon_Previous",
		"Text_BG","Music_BG"};
	char* I5_bmps[NUM_BMPS - 1] = {
	"Main_BG","Main_Icon_Movie","Main_Icon_Music",
		"Main_Icon_EBook","Main_Icon_About","Main_Icon_Help",
		"File_BG","File_Icon_Folder","File_Icon_Previous",
		"Text_BG","Music_BG"
	};

	// Icon names for offsets in ini file
	char* icons[NUM_OFFSETS] = {"Movie", "Music", "Image", "Game", "EBook",
		"Help", "FileText"};
	char* I5_icons[NUM_OFFSETS - 1] = {"Movie", "Music", "EBook",
		"About", "Help", "FileText"
	};

	// Names of text types for coloring
	char* texts[NUM_TEXT_COLORS] = {"Menu", "Highlight", "Text", "Reserved"};


	char  workDir[2000];
	char* backImgIniName = "BackImg.ini";
	CString backImgFilePath;
	char* backImgName = "BACKIMG";

	// File Pointers
	FILE* backImgFile;

	// Bitmap Data
	unsigned char* bmpData[NUM_BMPS];
	unsigned char* bmpPal[NUM_BG];

	// Error Codes
	int errorLevel = 0;

	// Firmware Version
	bool oldFirmware = false;

	// Config Data
	INI_BMP bmpInfo [NUM_BMPS];
	RGBN textColor [NUM_TEXT_COLORS];
	INI_OFFSET iconPos [NUM_OFFSETS];
	unsigned short showHelp;


	// temporary variables
	int i, j;
	CString tempString;

	// Startup information
	printf("GBAMP Skinner. Written by Chishm. Version: 1.21.\n");
	GetCurrentDirectory(2000,workDir);
	
	// Open and read ini file
	backImgFilePath = workDir ;
	backImgFilePath += "\\";
	backImgFilePath += backImgIniName;
	CINI ini;
	ini.setINIFileName (backImgFilePath);
	if (ini.sectionExists("FileVersion")){
		tempString = ini.getKeyValue("IniVersion","FileVersion");
		if (atoi(tempString) != iniVersion){
			printf("Warning: Ini file is not compatible with this version.\n");
			(errorLevel < 1) ? errorLevel = 1 : errorLevel ;
		}
	} else {
		printf("Ini file not found.\n");
		(errorLevel < 2) ? errorLevel = 2 : errorLevel ;
	}
	
	// Check which version of firmware this is for
	tempString = ini.getKeyValue("FirmwareVersion","FileVersion");
	tempString.MakeLower();
	if (tempString == oldFirmwareVersion) {
		oldFirmware = true;
	} else if (tempString == firmwareVersion) {
		oldFirmware = false;
	} else {
		printf ("Warning: Only firmware version e17 is supported.\n");
		(errorLevel < 1) ? errorLevel = 1 : errorLevel = errorLevel ;
		oldFirmware = false;		// Assume its meant for new firmware
	}

	// Read in bitmap data
	for (i = 0; i < (oldFirmware ? NUM_BMPS - 1 : NUM_BMPS); i++){
		bmpInfo[i] = iniGetBmpInfo ((oldFirmware ? I5_bmps[i] : bmps[i]), "Bitmaps", ini);
		bmpData[i] = LoadBmpData (bmpInfo[i]);
		if (bmpData[i] == NULL)	{
			(errorLevel < 2) ? errorLevel = 2 : errorLevel ;
		} else {
			for (j=0; j < bmpInfo[i].width * bmpInfo[i].height; j++){
				if (bmpData[i][j] >= PALETTE_RESERVED) {
					printf("Warning: %s uses reserved palette color %d.\n", bmps[i], bmpData[i][j]);
					(errorLevel < 1) ? errorLevel = 1 : errorLevel ;
					break;
				}
			}
		}
	}
	
	// Read in palette data.
	if (!oldFirmware){
		bmpPal[0] = LoadBmpPalette(bmpInfo[0].filename);
		bmpPal[1] = LoadBmpPalette(bmpInfo[7].filename);
		bmpPal[2] = LoadBmpPalette(bmpInfo[10].filename);
		bmpPal[3] = LoadBmpPalette(bmpInfo[11].filename);
	} else{
		bmpPal[0] = LoadBmpPalette(bmpInfo[0].filename);
		bmpPal[1] = LoadBmpPalette(bmpInfo[6].filename);
		bmpPal[2] = LoadBmpPalette(bmpInfo[9].filename);
		bmpPal[3] = LoadBmpPalette(bmpInfo[10].filename);
		
	}
	// Check mainscreen uses the same palette for all images
	for (i = 1; i < (oldFirmware ? 6:7); i++){
		if (CompareBmpPalette(bmpInfo[0].filename, bmpInfo[i].filename) != true){
			(errorLevel < 1) ? errorLevel = 1 : errorLevel ;
		}
		
	}

	// Check that file select uses the same palette for all images
	for (i = (oldFirmware ? 7:8); i < (oldFirmware ? 9:10); i++){
		if (CompareBmpPalette(bmpInfo[(oldFirmware ? 6:7)].filename, bmpInfo[i].filename) != true){
			(errorLevel <1) ? errorLevel = 1 : errorLevel ;
		}
	}


	// Read in color of text
	for (i = 0; i < NUM_TEXT_COLORS; i++) {
		textColor[i] = iniGetRGB(texts[i], "Color", ini);
	}

	// Read in icon offsets
	for (i = 0; i < (oldFirmware ? NUM_OFFSETS - 1 : NUM_OFFSETS); i++){
		iconPos[i] = iniGetOffset((oldFirmware ? I5_icons[i] : icons[i]), "MenuIconPostion", ini);
	}

	// Read in flags
	tempString = ini.getKeyValue("ShowHelpText","Flags");
	tempString.MakeLower();
	if (tempString == "true") {
		showHelp = 0x000A;
	} else {
		showHelp = 0x0000;
	}

	// Open BACKIMG file
	backImgFile = fopen (backImgName, "wb");
	if (backImgFile == NULL) {
		printf ("Unable to write to %s.\n", backImgName);
		(errorLevel < 2) ? errorLevel = 2 : errorLevel ;
	}
	
	if (errorLevel < 2) {
		
		// ***Write BACKIMG file***
		printf("Writing %s file.\n",backImgName);
		//Palettes and images first, in strange order due to way file is organised
		
		if (oldFirmware == false){
			// Version e17
			fwrite (bmpPal[0], PALETTE_LENGTH,1,backImgFile);
			for ( i = 0; i < 6; i++){
				fwrite (bmpData[i], bmpInfo[i].height * bmpInfo[i].width, 1, backImgFile);
			}
			fwrite (bmpPal[1], PALETTE_LENGTH,1,backImgFile);
			for ( i = 7; i < 10; i++){
				fwrite (bmpData[i], bmpInfo[i].height * bmpInfo[i].width, 1, backImgFile);
			}
			fwrite (bmpPal[2], PALETTE_LENGTH,1,backImgFile);
			i = 10;
			fwrite (bmpData[i], bmpInfo[i].height * bmpInfo[i].width, 1, backImgFile);
			fwrite (bmpPal[3], PALETTE_LENGTH,1,backImgFile);
			i = 11;
			fwrite (bmpData[i], bmpInfo[i].height * bmpInfo[i].width, 1, backImgFile);
			i = 6;
			fwrite (bmpData[i], bmpInfo[i].height * bmpInfo[i].width, 1, backImgFile);
		} else {
			// Version e10
			fwrite (bmpPal[0], PALETTE_LENGTH,1,backImgFile);
			for ( i = 0; i < 6; i++){
				fwrite (bmpData[i], bmpInfo[i].height * bmpInfo[i].width, 1, backImgFile);
			}
			fwrite (bmpPal[1], PALETTE_LENGTH,1,backImgFile);
			for ( i = 6; i < 9; i++){
				fwrite (bmpData[i], bmpInfo[i].height * bmpInfo[i].width, 1, backImgFile);
			}
			fwrite (bmpPal[2], PALETTE_LENGTH,1,backImgFile);
			i = 9;
			fwrite (bmpData[i], bmpInfo[i].height * bmpInfo[i].width, 1, backImgFile);
			fwrite (bmpPal[3], PALETTE_LENGTH,1,backImgFile);
			i = 10;
			fwrite (bmpData[i], bmpInfo[i].height * bmpInfo[i].width, 1, backImgFile);
		}

		// Text Colors
		for (i = 0; i < NUM_TEXT_COLORS; i++){
			fwrite(&textColor[i], sizeof(RGBN), 1, backImgFile);
		}
		
		// Icon offsets
		for ( i=0; i < NUM_OFFSETS; i++){
			fwrite(&iconPos[i],sizeof(INI_OFFSET),1, backImgFile);
		}
		
		// Flags
		fwrite(&showHelp, sizeof(short), 1, backImgFile);

	} else {
		printf ("Could not write %s due to errors.\n", backImgName);
	}  // ErrorLevel < 2

	// Close BACKIMG file
	fclose (backImgFile);
	
	// Free memory used
	for (i = 0; i < NUM_BMPS; i++){
		free(bmpData[i]);
	}
	for (i = 0; i < NUM_BG; i++){
		free(bmpPal[i]);
	}

	if (errorLevel == 0) {
		printf("BACKIMG skin successfully created.\n");
	}

	return errorLevel;

}
